/**
 * This bit indicates whether a number is considered a system error code.
 * If yes, the system errno is just the result of clearing this bit from
 * the given number.
 */
#define SYSTEM_ERROR_BIT 30

/** Check whether the system error bit is set. */
#define IS_SYSTEM_ERROR(num) (!!((num) & (1 << SYSTEM_ERROR_BIT)))

/** Set the system error bit for the given number. */
#define ERRNO_TO_ERROR(num) ((num) | (1 << SYSTEM_ERROR_BIT))

#define ALL_ERRORS \
	_ERROR(SUCCESS, "success") \
	_ERROR(SYNTAX, "syntax error") \
	_ERROR(LOOP_COMPLETE, "loop complete") \
	_ERROR(HASH_TABLE_OVERFLOW, "hash table too small") \
	_ERROR(BAD_UID, "uid not found in hash table") \
	_ERROR(ATOI_OVERFLOW, "value too large") \
	_ERROR(STRTOLL, "unknown strtoll error") \
	_ERROR(ATOI_NO_DIGITS, "no digits found in string") \
	_ERROR(ATOI_JUNK_AT_END, "further characters after number") \
	_ERROR(EMPTY, "file empty") \
	_ERROR(MMAP, "mmap error") \
	_ERROR(OSL, "osl error") \
	_ERROR(SIGNAL_SIG_ERR, "signal() returned SIG_ERR") \
	_ERROR(OUTPUT, "error writing output") \
	_ERROR(MALFORMED_FORMAT, "malformed format string") \
	_ERROR(BAD_ALIGN_SPEC, "bad alignment specifier") \
	_ERROR(TRAILING_GARBAGE, "trailing garbage after specifier") \
	_ERROR(UNIT, "no unit allowed here") \
	_ERROR(BAD_UNIT, "invalid unit specifier") \
	_ERROR(BAD_ATOM, "invalid atom") \
	_ERROR(BAD_OUTPUT_ARG, "invalid name for output") \


/**
 * This is temporarily defined to expand to its first argument (prefixed by
 * 'E_') and gets later redefined to expand to the error text only
 */
#define _ERROR(err, msg) E_ ## err,

enum error_codes {
	ALL_ERRORS
};
#undef _ERROR
#define _ERROR(err, msg) msg,
#define DEFINE_ERRLIST char *adu_errlist[] = {ALL_ERRORS}

extern int osl_errno;
extern char *adu_errlist[];


/**
 * adu's version of strerror(3).
 *
 * \param num The error number.
 *
 * \return The error text of \a num.
 */
static inline const char *adu_strerror(int num)
{
	assert(num > 0);
	if (num == E_OSL) {
		assert(osl_errno > 0);
		return osl_strerror((osl_errno));
	}
	if (IS_SYSTEM_ERROR(num))
		return strerror((num) & ((1 << SYSTEM_ERROR_BIT) - 1));
	return adu_errlist[num];
}

/**
 * Wrapper for osl library calls.
 *
 * This should be used for all calls to osl functions that return an osl error
 * code. It changes the return value to \p -E_OSL appropriately so that it can
 * be used for printing the correct error message.
 *
 * \return \a ret if \a ret >= 0, \p -E_OSL otherwise.
 */
static inline int osl(int ret)
{
	if (ret >= 0)
		return ret;
	osl_errno = -ret;
	return -E_OSL;
}
