/*
 * Copyright (C) 2008 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file select.c The select mode of adu. */

#include <dirent.h> /* readdir() */
#include "format.h"
#include "adu.h"
#include "gcc-compat.h"
#include "cmdline.h"
#include "fd.h"
#include "string.h"
#include "error.h"
#include "user.h"
#include "select.cmdline.h"

/* global list */
#define GLOBAL_LIST_ATOMS \
	ATOM(size, SIZE) \
	ATOM(files, COUNT) \
	ATOM(dirname, STRING) \

#define ATOM(x, y) { .name = #x, .type = AT_ ## y},
struct atom global_list_atoms[] = {
	GLOBAL_LIST_ATOMS
	{.name = NULL}
};
#undef ATOM
#define ATOM(x, y) gla_ ## x,
enum global_list_atoms {GLOBAL_LIST_ATOMS};
#undef ATOM

/* global summary */
#define GLOBAL_SUMMARY_ATOMS \
	ATOM(dirs, COUNT) \
	ATOM(files, COUNT) \
	ATOM(size, SIZE)

#define ATOM(x, y) { .name = #x, .type = AT_ ## y},
struct atom global_summary_atoms[] = {
	GLOBAL_SUMMARY_ATOMS
	{.name = NULL}
};
#undef ATOM
#define ATOM(x, y) gsa_ ## x,
enum global_summary_atoms {GLOBAL_SUMMARY_ATOMS};
#undef ATOM

/* user list */
#define USER_LIST_ATOMS \
	ATOM(pw_name, STRING) \
	ATOM(uid, ID) \
	ATOM(size, SIZE) \
	ATOM(files, COUNT) \
	ATOM(dirname, STRING) \

#define ATOM(x, y) { .name = #x, .type = AT_ ## y},
struct atom user_list_atoms[] = {
	USER_LIST_ATOMS
	{.name = NULL}
};
#undef ATOM
#define ATOM(x, y) ula_ ## x,
enum user_list_atoms {USER_LIST_ATOMS};
#undef ATOM

/* user summary */
#define USER_SUMMARY_ATOMS \
	ATOM(pw_name, STRING) \
	ATOM(uid, ID) \
	ATOM(dirs, COUNT) \
	ATOM(files, COUNT) \
	ATOM(size, SIZE)

#define ATOM(x, y) { .name = #x, .type = AT_ ## y},
struct atom user_summary_atoms[] = {
	USER_SUMMARY_ATOMS
	{.name = NULL}
};
#undef ATOM
#define ATOM(x, y) usa_ ## x,
enum user_summary_atoms {USER_SUMMARY_ATOMS};
#undef ATOM

struct global_list_info {
	uint32_t count;
	int ret;
	int osl_errno;
	struct format_info *fi;
};

struct global_summary_info {
	/** Global dir count. */
	uint64_t num_dirs;
	/** Global files count. */
	uint64_t num_files;
	/** Global bytes count. */
	uint64_t num_bytes;
	int ret;
	int osl_errno;
};

struct user_list_info {
	uint32_t count;
	struct user_info *ui;
	struct format_info *fi;
	int ret;
	int osl_errno;
};

struct user_summary_info {
	struct user_info *ui;
	int ret;
	int osl_errno;
};


static FILE *output_file;

__printf_1_2 static int output(const char const *fmt, ...)
{
	va_list argp;
	int ret;

	va_start(argp, fmt);
	ret = vfprintf(output_file, fmt, argp);
	va_end(argp);
	return ret < 0? -E_OUTPUT : 1;
}

static int get_dir_name_by_number(uint64_t *dirnum, char **name)
{
	char *result = NULL, *tmp;
	struct osl_row *row;
	uint64_t val = *dirnum;
	struct osl_object obj;
	int ret;
	char *pfx;

again:
	obj.data = &val;
	obj.size = sizeof(val);
	ret = osl(osl_get_row(dir_table, DT_NUM, &obj, &row));
	if (ret < 0)
		goto out;
	ret = osl(osl_get_object(dir_table, row, DT_PARENT_NUM, &obj));
	if (ret < 0)
		goto out;
	val = *(uint64_t *)obj.data;
	ret = osl(osl_get_object(dir_table, row, DT_NAME, &obj));
	if (ret < 0)
		goto out;
	pfx = (select_conf.print_base_dir_given || val)? (char *)obj.data :  ".";
	tmp = make_message("%s/%s", pfx, result? result : "");
	free(result);
	result = tmp;
	if (val)
		goto again;
out:
	if (ret < 0) {
		free(result);
		*name = NULL;
	} else {
		assert(result);
		*name = result;
	}
	return ret;
}

static int get_dir_name_of_row(struct osl_row *dir_table_row, char **name)
{
	struct osl_object obj;
	int ret;

	*name = NULL;
	ret = osl(osl_get_object(dir_table, dir_table_row, DT_NUM, &obj));
	if (ret < 0)
		return ret;
	return get_dir_name_by_number((uint64_t *)obj.data, name);
}

static int get_dir_name_of_user_row(struct osl_row *user_table_row,
		struct user_info *ui, char **dirname)
{
	struct osl_object obj;
	int ret = osl(osl_get_object(ui->table, user_table_row,
		UT_DIR_NUM, &obj));

	if (ret < 0)
		return ret;
	return get_dir_name_by_number((uint64_t *)obj.data, dirname);
}

static int get_num_files_of_row(struct osl_row *row, uint64_t *num_files)
{
	struct osl_object obj;
	int ret = osl(osl_get_object(dir_table, row, DT_FILES, &obj));
	if (ret < 0)
		return ret;
	*num_files = *(uint64_t *)obj.data;
	return 1;
}

static int get_num_user_files(struct osl_row *row, struct user_info *ui,
		uint64_t *num_files)
{
	struct osl_object obj;
	int ret = osl(osl_get_object(ui->table, row, UT_FILES, &obj));

	if (ret < 0)
		return ret;
	*num_files = *(uint64_t *)obj.data;
	return 1;
}

static int get_num_bytes_of_row(struct osl_row *row, uint64_t *num_bytes)
{
	struct osl_object obj;
	int ret = osl(osl_get_object(dir_table, row, DT_BYTES, &obj));
	if (ret < 0)
		return ret;
	*num_bytes = *(uint64_t *)obj.data;
	return 1;
}

static int get_num_user_bytes(struct osl_row *row, struct user_info *ui,
		uint64_t *num_bytes)
{
	struct osl_object obj;
	int ret = osl(osl_get_object(ui->table, row, UT_BYTES, &obj));

	if (ret < 0)
		return ret;
	*num_bytes = *(uint64_t *)obj.data;
	return 1;
}

static int check_loop_return(int ret, int loop_ret, int loop_osl_errno)
{
	if (ret >= 0)
		return ret;
	assert(ret == -E_OSL);
	if (osl_errno != E_OSL_LOOP)
		/* error not caused by loop function returning negative. */
		return ret;
	assert(loop_ret < 0);
	if (loop_ret == -E_LOOP_COMPLETE) /* no error */
		return 1;
	if (loop_ret == -E_OSL) { /* osl error in loop function */
		assert(loop_osl_errno);
		osl_errno = loop_osl_errno;
	}
	return loop_ret;
}

static int adu_loop_reverse(struct osl_table *t, unsigned col_num, void *private_data,
		osl_rbtree_loop_func *func, int *loop_ret, int *loop_osl_errno)
{
	int ret = osl(osl_rbtree_loop_reverse(t, col_num, private_data, func));
	return check_loop_return(ret, *loop_ret, *loop_osl_errno);
}

static int global_summary_loop_function(struct osl_row *row, void *data)
{
	struct global_summary_info *gsi = data;
	int ret;
	uint64_t num;

	ret = get_num_files_of_row(row, &num);
	if (ret < 0)
		goto err;
	gsi->num_files += num;

	ret = get_num_bytes_of_row(row, &num);
	if (ret < 0)
		goto err;
	gsi->num_bytes += num;
	gsi->num_dirs++;
	return 1;
err:
	gsi->ret = ret;
	gsi->osl_errno = (ret == -E_OSL)? osl_errno : 0;
	return ret;
}

static int print_global_summary(struct format_info *fi)
{
	int ret;
	char *buf;
	struct global_summary_info gsi = {.num_dirs = 0};

	union atom_value values[] = {
		[gsa_dirs] = {.num_value = 0ULL},
		[gsa_files] = {.num_value =  0ULL},
		[gsa_size] = {.num_value =  0ULL}
	};

	ret = adu_loop_reverse(dir_table, DT_BYTES, &gsi,
		global_summary_loop_function, &gsi.ret, &gsi.osl_errno);
	if (ret < 0)
		return ret;
	values[gsa_dirs].num_value = (long long unsigned)gsi.num_dirs;
	values[gsa_files].num_value = (long long unsigned)gsi.num_files;
	values[gsa_size].num_value = (long long unsigned)gsi.num_bytes;
	if (!select_conf.no_headers_given) {
		ret = output("Global summary\n");
		if (ret < 0)
			return ret;
	}
	buf = format_items(fi, values);
	ret = output("%s", buf);
	free(buf);
	return ret;
}

static int user_summary_loop_function(struct osl_row *row, void *data)
{
	struct user_summary_info *usi = data;
	uint64_t num;
	int ret;

	ret = get_num_user_files(row, usi->ui, &num);
	if (ret < 0)
		goto err;
	usi->ui->files += num;
	ret = get_num_user_bytes(row, usi->ui, &num);
	if (ret < 0)
		goto err;
	usi->ui->bytes += num;
	usi->ui->dirs++;
	return 1;
err:
	usi->ret = ret;
	usi->osl_errno = (ret == -E_OSL)? osl_errno : 0;
	return ret;
}

static int compute_user_summary(struct user_info *ui, __a_unused void *data)
{
	struct user_summary_info usi = {.ui = ui};

	return adu_loop_reverse(ui->table, UT_BYTES, &usi, user_summary_loop_function,
		&usi.ret, &usi.osl_errno);
}

static int print_user_summary_line(struct user_info *ui, void *data)
{
	struct format_info *fi = data;
	union atom_value values[] = {
		[usa_pw_name] = {.string_value = ui->pw_name?
			ui->pw_name : "?"},
		[usa_uid] = {.num_value = (long long unsigned)ui->uid},
		[usa_dirs] = {.num_value = (long long unsigned)ui->dirs},
		[usa_files] = {.num_value = (long long unsigned)ui->files},
		[usa_size] = {.num_value =  (long long unsigned)ui->bytes}
	};
	char *buf;
	int ret;

	buf = format_items(fi, values);
	ret = output("%s", buf);
	free(buf);
	return ret;
}

static int name_comp(const void *a, const void *b)
{
	char *x = ((struct user_info *)a)->pw_name;
	char *y = ((struct user_info *)b)->pw_name;

	if (!x)
		return 1;
	if (!y)
		return -1;
	return strcmp(x, y);
}

static int uid_comp(const void *a, const void *b)
{
	return -NUM_COMPARE(((struct user_info *)a)->uid,
		((struct user_info *)b)->uid);
}

static int dir_count_comp(const void *a, const void *b)
{
	return NUM_COMPARE(((struct user_info *)a)->dirs,
		((struct user_info *)b)->dirs);
}

static int file_count_comp(const void *a, const void *b)
{
	return NUM_COMPARE(((struct user_info *)a)->files,
		((struct user_info *)b)->files);
}

static int size_comp(const void *a, const void *b)
{
	return NUM_COMPARE(((struct user_info *)a)->bytes,
		((struct user_info *)b)->bytes);
}

static int print_user_summary(struct format_info *fi)
{
	/*
	 * The comparators for sorting the user summary.
	 *
	 * This is an array of pointers to functions taking two constant void *
	 * pointers and returning an int.
	 */
	static int (*summary_comparators[])(const void *, const void *) = {
		[user_summary_sort_arg_name] = name_comp,
		[user_summary_sort_arg_uid] = uid_comp,
		[user_summary_sort_arg_dir_count] = dir_count_comp,
		[user_summary_sort_arg_file_count] = file_count_comp,
		[user_summary_sort_arg_size] = size_comp,
	};

	if (!select_conf.no_headers_given) {
		int ret = output("User summary\n");
		if (ret < 0)
			return ret;
	}
	int ret = for_each_admissible_user(compute_user_summary, fi);
	if (ret < 0)
		return ret;
	sort_hash_table(summary_comparators[select_conf.user_summary_sort_arg]);
	return for_each_admissible_user(print_user_summary_line, fi);
}

static int user_list_loop_function(struct osl_row *row, void *data)
{
	struct user_list_info *uli = data;
	union atom_value values[] = {
		[ula_pw_name] = {.string_value = uli->ui->pw_name?
			uli->ui->pw_name : "?"},
		[ula_uid] = {.num_value = (long long unsigned)uli->ui->uid},
		[ula_files] = {.num_value = 0ULL},
		[ula_size] = {.num_value =  0ULL},
		[ula_dirname] = {.string_value = NULL}
	};
	uint64_t num;
	int ret;
	char *dirname, *buf;

	check_signals();
	ret = -E_LOOP_COMPLETE;
	if (!uli->count)
		goto err;

	ret = get_num_user_files(row, uli->ui, &num);
	if (ret < 0)
		goto err;
	values[ula_files].num_value = num;

	ret = get_num_user_bytes(row, uli->ui, &num);
	if (ret < 0)
		goto err;
	values[ula_size].num_value = num;

	ret = get_dir_name_of_user_row(row, uli->ui, &dirname);
	if (ret < 0)
		goto err;
	values[ula_dirname].string_value = dirname;

	buf = format_items(uli->fi, values);
	free(dirname);
	ret = output("%s", buf);
	free(buf);
	if (ret < 0)
		goto err;
	uli->count--;
	return ret;
err:
	uli->ret = ret;
	uli->osl_errno = (ret == -E_OSL)? osl_errno : 0;
	return ret;
}

static int print_user_list(struct user_info *ui, void *data)
{
	struct format_info *fi = data;
	int ret;
	enum user_table_columns sort_column = UT_BYTES;
	struct user_list_info uli = {
		.ui = ui,
		.fi = fi,
		.count = select_conf.limit_arg
	};

	if (select_conf.list_sort_arg == list_sort_arg_file_count)
		sort_column = UT_FILES;

	if (!select_conf.no_headers_given) {
		ret = output("%s (uid %u)\n",
			ui->pw_name? ui->pw_name : "?", (unsigned)ui->uid);
		if (ret < 0)
			return ret;
	}
	return adu_loop_reverse(ui->table, sort_column, &uli, user_list_loop_function,
		&uli.ret, &uli.osl_errno);
}

static int global_list_loop_function(struct osl_row *row, void *data)
{
	struct global_list_info *gli = data;
	union atom_value values[] = {
		[gla_size] = {.num_value = 0ULL},
		[gla_files] = {.num_value =  0ULL},
		[gla_dirname] = {.string_value = NULL}
	};
	uint64_t num_files, num_bytes;
	char *dirname, *buf;
	int ret;

	check_signals();
	ret = -E_LOOP_COMPLETE;
	if (!gli->count)
		goto err;

	ret = get_num_files_of_row(row, &num_files);
	if (ret < 0)
		goto err;
	values[gla_files].num_value = (long long unsigned)num_files;

	ret = get_num_bytes_of_row(row, &num_bytes);
	if (ret < 0)
		goto err;
	values[gla_size].num_value = (long long unsigned)num_bytes;

	ret = get_dir_name_of_row(row, &dirname);
	if (ret < 0)
		goto err;
	values[gla_dirname].string_value = dirname;

	buf = format_items(gli->fi, values);
	free(dirname);
	ret = output("%s", buf);
	free(buf);
	if (ret < 0)
		goto err;
	if (gli->count > 0)
		gli->count--;
	return ret;
err:
	gli->ret = ret;
	gli->osl_errno = (ret == -E_OSL)? osl_errno : 0;
	return -1;
}

static int print_global_list(struct format_info *fi)
{
	int ret;
	enum dir_table_columns sort_column = DT_BYTES;
	struct global_list_info gli = {
		.fi = fi,
		.count = select_conf.limit_arg
	};

	if (!select_conf.no_headers_given) {
		ret = output("Global list\n");
		if (ret < 0)
			return ret;
	}
	if (select_conf.list_sort_arg == list_sort_arg_file_count)
		sort_column = DT_FILES;
	return adu_loop_reverse(dir_table, sort_column, &gli,
		global_list_loop_function, &gli.ret, &gli.osl_errno);
}

static int print_statistics(struct format_info *fi)
{
	switch (select_conf.select_mode_arg) {
		case select_mode_arg_global_list:
			return print_global_list(fi);
		case select_mode_arg_global_summary:
			return print_global_summary(fi);
		case select_mode_arg_user_list:
			return for_each_admissible_user(print_user_list, fi);
		case select_mode_arg_user_summary:
			return print_user_summary(fi);
	};
	ERROR_LOG("bad select mode\n");
	return -ERRNO_TO_ERROR(EINVAL);
}

static int open_pipe(char *path)
{
	int p[2], ret, argc;
	char **argv;

	ret = pipe(p);
	if (ret < 0)
		return ERRNO_TO_ERROR(errno);
	ret = fork();
	if (ret < 0)
		return ERRNO_TO_ERROR(errno);
	if (ret) { /* parent */
		DEBUG_LOG("created process %d\n", ret);
		close(p[0]);
		output_file = fdopen(p[1], "w");
		if (!output_file)
			return ERRNO_TO_ERROR(errno);
		return 1;
	}
	close(p[1]);
	if (p[0] != STDIN_FILENO)
		dup2(p[0], STDIN_FILENO);
	DEBUG_LOG("executing %s\n", path);
	argc = split_args(path, &argv, " \t");
	execvp(argv[0], argv);
	ERROR_LOG("error executing %s: %s\n", path,
		adu_strerror(ERRNO_TO_ERROR(errno)));
	_exit(EXIT_FAILURE);
}

static int open_output_stream(void)
{
	char *p;
	int ret, flags = O_WRONLY | O_CREAT;

	if (!select_conf.output_given)
		goto use_stdout;
	p = select_conf.output_arg;
	switch (p[0]) {
	case '\0': /* empty string */
		goto bad_output_arg;
	case '-':
		if (!p[1]) /* "-" means stdout */
			goto use_stdout;
		/* string starting with a dash */
		flags |= O_EXCL;
		goto open_file;
	case '>':
		if (!p[1]) /* ">" is invalid */
			goto bad_output_arg;
		if (p[1] != '>') {
			p++;
			flags |= O_TRUNC;
			goto open_file;
		}
		/* string starting with ">>" */
		if (!p[2]) /* ">>" is invalid */
			goto bad_output_arg;
		flags |= O_APPEND;
		p += 2;
		goto open_file;
	case '|':
		if (!p[1]) /* "|" is invalid */
			goto bad_output_arg;
		p++;
		return open_pipe(p);
	default: /* args starts with no magic character */
		flags |= O_EXCL;
		goto open_file;
	}
use_stdout:
	output_file = stdout;
	return 1;
bad_output_arg:
	output_file = NULL;
	return -E_BAD_OUTPUT_ARG;
open_file:
	/*
	 * glibc's 'x' mode to fopen is not portable, so use open() and
	 * fdopen().
	 */
	ret = open(p, flags, 0644);
	if (ret < 0)
		return -ERRNO_TO_ERROR(errno);
	output_file = fdopen(ret, "w");
	if (!output_file)
		return -ERRNO_TO_ERROR(errno);
	return 1;
}

int run_select_query(struct uid_range *admissible_uids, struct format_info *fi)
{
	int ret = open_output_stream();

	if (ret < 0)
		goto out;
	ret = open_dir_table(0);
	if (ret < 0)
		goto out;
	check_signals();
	ret = read_uid_file(admissible_uids);
	if (ret < 0)
		goto out;
	check_signals();
	ret = print_statistics(fi);
out:
	close_all_tables();
	if (output_file && output_file != stdout) {
		fclose(output_file);
		output_file = NULL;
	}
	return ret;
}

#define GLOBAL_LIST_DFLT_FMT "%(size:r:8) %(files:r:8) %(dirname)\n"
#define GLOBAL_SUMMARY_DFLT_FMT "#directories: %(dirs), #files: %(files), size: %(size)\n\n"
#define USER_LIST_DFLT_FMT "%(size:r:5) %(files:r:5) %(dirname)\n"
#define USER_SUMMARY_DFLT_FMT "%(pw_name:l:16) %(uid:r:5) %(dirs:r:5) %(files:r:5) %(size:r:5)\n"

static int setup_format_string(char *fmt, struct format_info **fi)
{
	struct atom *atoms;

	if (!fmt)
		INFO_LOG("using default format string\n");
	switch (select_conf.select_mode_arg) {
	case select_mode_arg_global_list:
		if (!fmt)
			fmt = GLOBAL_LIST_DFLT_FMT;
		atoms = global_list_atoms;
		break;
	case select_mode_arg_global_summary:
		if (!fmt)
			fmt = GLOBAL_SUMMARY_DFLT_FMT;
		atoms = global_summary_atoms;
		break;
	case select_mode_arg_user_list:
		if (!fmt)
			fmt = USER_LIST_DFLT_FMT;
		atoms = user_list_atoms;
		break;
	case select_mode_arg_user_summary:
		if (!fmt)
			fmt = USER_SUMMARY_DFLT_FMT;
		atoms = user_summary_atoms;
		break;
	default:
		ERROR_LOG("bad select mode\n");
		return -ERRNO_TO_ERROR(EINVAL);
	};
	INFO_LOG("format string: %s\n", fmt);
	return parse_format_string(fmt, atoms, fi);
}

/* return: < 0: error, >0: OK, == 0: help given */
int parse_select_options(char *string, struct select_cmdline_parser_params *params,
		struct uid_range **admissible_uids, struct format_info **fi)
{
	int ret, num_uid_ranges;
	const char **line;
	char *fmt = NULL;

	if (string) {
		int argc;
		char **argv;

		ret = create_argv(string, &argv);
		if (ret < 0)
			return ret;
		argc = ret;
		ret = select_cmdline_parser_ext(argc, argv, &select_conf, params);
		free_argv(argv);
		if (ret)
			return -E_SYNTAX;
		if (select_conf.help_given || select_conf.detailed_help_given)
			goto help;
		fmt = select_conf.format_arg;
	}
	ret = parse_uid_arg(select_conf.uid_arg, admissible_uids);
	if (ret < 0)
		return ret;
	num_uid_ranges = ret;
	ret = append_users(select_conf.user_arg, select_conf.user_given,
		admissible_uids, num_uid_ranges);
	if (ret < 0)
		return ret;
	return setup_format_string(fmt, fi);
help:
	line = select_conf.detailed_help_given?
		select_args_info_detailed_help : select_args_info_help;
	if (!output_file)
		output_file = stdout;
	for (; *line; line++) {
		ret = output("%s\n", *line);
		if (ret < 0)
			return ret;
	}
	return 0;
}

int com_select(void)
{
	struct uid_range *admissible_uids = NULL;
	int ret;
	struct format_info *fi;
	struct select_cmdline_parser_params params = {
		.override = 1,
		.initialize = 1,
		.check_required = 1,
		.check_ambiguity = 1,
		.print_errors = 1
	};

	select_cmdline_parser_init(&select_conf);
	ret = parse_select_options(conf.select_options_arg, &params,
		&admissible_uids, &fi);
	if (ret <= 0) /* do not run query if help was given */
		return ret;
	ret = run_select_query(admissible_uids, fi);
	free_format_info(fi);
	return ret;
}
